# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from .meta import (
    Datum,
    InConverter,
    OutConverter,
    _BaseConverter,
    _ConverterMeta,
    get_binding_registry,
)
from .sdkType import SdkType
from .web import (
    HttpV2FeatureChecker,
    ModuleTrackerMeta,
    RequestSynchronizer,
    RequestTrackerMeta,
    ResponseLabels,
    ResponseTrackerMeta,
    WebApp,
    WebServer,
)

__all__ = [
    "Datum",
    "_ConverterMeta",
    "_BaseConverter",
    "InConverter",
    "OutConverter",
    "SdkType",
    "get_binding_registry",
    "ModuleTrackerMeta",
    "RequestTrackerMeta",
    "ResponseTrackerMeta",
    "HttpV2FeatureChecker",
    "ResponseLabels",
    "WebServer",
    "WebApp",
    "RequestSynchronizer",
]

__version__ = "1.0.0b2"



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import inspect
import json
import re
from abc import ABC
from enum import Enum
from typing import Any, Callable, Dict, List, Optional

from . import meta

SNAKE_CASE_RE = re.compile(r"^([a-zA-Z]+\d*_|_+[a-zA-Z\d])\w*$")
WORD_RE = re.compile(r"^([a-zA-Z]+\d*)$")


class StringifyEnum(Enum):
    """This class output name of enum object when printed as string."""

    def __str__(self):
        return str(self.name)


class StringifyEnumJsonEncoder(json.JSONEncoder):
    def default(self, o):
        return str(o)


class BuildDictMeta(type):
    def __new__(mcs, name, bases, dct):
        """BuildDictMeta will apply to every binding.
        It will apply :meth:`add_to_dict` decorator to :meth:`__init__` of
        every binding class to collect list of params to include in building
        json dictionary which corresponds to function.json in legacy app.
        It will also apply :meth:`skip_none` to :meth:`get_dict_repr` to
        enable json dictionary generated for every binding has non-empty
        value fields. It is needed for enabling binding param optionality.
        """
        cls = super().__new__(mcs, name, bases, dct)
        setattr(cls, "__init__", cls.add_to_dict(getattr(cls, "__init__")))
        setattr(cls, "get_dict_repr", cls.skip_none(getattr(cls, "get_dict_repr")))
        return cls

    @staticmethod
    def skip_none(func):
        def wrapper(*args, **kw):
            res = func(*args, **kw)
            return BuildDictMeta.clean_nones(res)

        return wrapper

    @staticmethod
    def add_to_dict(func: Callable[..., Any]):
        def wrapper(*args, **kwargs):
            if args is None or len(args) == 0:
                raise ValueError(
                    f"{func.__name__} has no args. Please ensure func is an "
                    f"object method."
                )

            func(*args, **kwargs)

            self = args[0]

            init_params = list(inspect.signature(func).parameters.keys())
            init_params.extend(list(kwargs.keys()))
            for key in kwargs.keys():
                if not hasattr(self, key):
                    setattr(self, key, kwargs[key])

            setattr(self, "init_params", init_params)

        return wrapper

    @staticmethod
    def clean_nones(value):
        """
        Recursively remove all None values from dictionaries and lists,
        and returns
        the result as a new dictionary or list.
        """
        if isinstance(value, list):
            return [BuildDictMeta.clean_nones(x) for x in value if x is not None]
        elif isinstance(value, dict):
            return {
                key: BuildDictMeta.clean_nones(val)
                for key, val in value.items()
                if val is not None
            }
        else:
            return value


# Enums
class BindingDirection(StringifyEnum):
    """Direction of the binding used in function.json"""

    IN = 0
    """Input binding direction."""
    OUT = 1
    """Output binding direction."""
    INOUT = 2
    """Some bindings support a special binding direction. """


class DataType(StringifyEnum):
    """Data type of the binding used in function.json"""

    """Parse binding argument as undefined."""
    UNDEFINED = 0
    """Parse binding argument as string."""
    STRING = 1
    """Parse binding argument as binary."""
    BINARY = 2
    """Parse binding argument as stream."""
    STREAM = 3


class Binding(ABC):
    """Abstract binding class which captures common attributes and
    functions. :meth:`get_dict_repr` can auto generate the function.json for
    every binding, the only restriction is ***ENSURE*** __init__ parameter
    names of any binding class are snake case form of corresponding
    attribute in function.json when new binding classes are created.
    Ref: https://aka.ms/azure-function-binding-http"""

    EXCLUDED_INIT_PARAMS = {"self", "kwargs", "type", "data_type", "direction"}

    @staticmethod
    def get_binding_name() -> str:
        pass

    def __init__(
        self,
        name: str,
        direction: BindingDirection,
        data_type: Optional[DataType] = None,
        type: Optional[str] = None,
    ):  # NoQa
        # For natively supported bindings, get_binding_name is always
        # implemented, and for generic bindings, type is a required argument
        # in decorator functions.
        self.type = (
            self.get_binding_name() if self.get_binding_name() is not None else type
        )
        self.name = name
        self._direction = direction
        self._data_type = data_type
        self._dict = {
            "direction": self._direction,
            "dataType": self._data_type,
            "type": self.type,
        }

    @property
    def data_type(self) -> Optional[int]:
        return self._data_type.value if self._data_type else None

    @property
    def direction(self) -> int:
        return self._direction.value

    def get_dict_repr(binding, input_types):
        """Build a dictionary of a particular binding. The keys are camel
        cased binding field names defined in `init_params` list and
        :class:`Binding` class. \n
        This method is invoked in function :meth:`get_raw_bindings` of class
        :class:`Function` to generate json dict for each binding.

        :return: Dictionary representation of the binding. Dict representation
        of the binding in the format:
        ((binding type, pytype), deferred bindings enabled)
        """
        params = list(dict.fromkeys(getattr(binding, "init_params", [])))
        binding_info = {}
        for p in params:
            if p not in Binding.EXCLUDED_INIT_PARAMS:
                binding._dict[to_camel_case(p)] = getattr(binding, p, None)

        if input_types.get(binding.name) is not None:
            pytype = input_types.get(binding.name).pytype
        else:
            pytype = None
        # Adding flag to signal to the host to send MBD object
        # 1. check if the binding is a supported type (blob, blobTrigger)
        # 2. check if the binding is an input binding
        # 3. check if the defined type is an SdkType
        if (
            binding.type in meta._ConverterMeta._bindings
            and binding.direction == 0
            and meta._ConverterMeta.check_supported_type(pytype)
        ):
            binding._dict["properties"] = {"SupportsDeferredBinding": True}
            binding_info = {binding.name: {pytype: "True"}}
        # if it isn't, we set the flag to false
        else:
            binding._dict["properties"] = {"SupportsDeferredBinding": False}
            binding_info = {binding.name: {pytype: "False"}}

        return binding._dict, binding_info


def to_camel_case(snake_case_str: str):
    if snake_case_str is None or len(snake_case_str) == 0:
        raise ValueError(f"Please ensure arg name {snake_case_str} is not empty!")

    if not is_snake_case(snake_case_str) and not is_word(snake_case_str):
        raise ValueError(
            f"Please ensure {snake_case_str} is a word or snake case "
            f"string with underscore as separator."
        )
    words = snake_case_str.split("_")
    return words[0] + "".join([ele.title() for ele in words[1:]])


def is_snake_case(input_string: str) -> bool:
    """
    Checks if a string is formatted as "snake case".
    A string is considered snake case when:
    - it's composed only by lowercase/uppercase letters and digits
    - it contains at least one underscore
    - it does not start with a number
    *Examples:*
    >>> is_snake_case('foo_bar_baz') # returns true
    >>> is_snake_case('foo') # returns false
    :param input_string: String to test.
    :return: True for a snake case string, false otherwise.
    """
    return SNAKE_CASE_RE.match(input_string) is not None


def is_word(input_string: str) -> bool:
    """
    Checks if a string is one word.
    A string is considered one word when:
    - it's composed only by lowercase/uppercase letters and digits
    - it does not start with a number
    *Examples:*
    >>> is_word('1foo') # returns false
    >>> is_word('foo_') # returns false
    >>> is_word('foo') # returns true
    :param input_string: String to test.
    :return: True for one word string, false otherwise.
    """
    return WORD_RE.match(input_string) is not None


def get_raw_bindings(indexed_function, input_types):
    binding_dict_repr = []
    bindings_logs = {}
    for b in indexed_function._bindings:
        dict_repr, logs = Binding.get_dict_repr(b, input_types)
        clean_dict_repr = BuildDictMeta.clean_nones(dict_repr)
        binding_dict_repr.append(
            json.dumps(clean_dict_repr, cls=StringifyEnumJsonEncoder)
        )
        bindings_logs.update(logs)
    return binding_dict_repr, bindings_logs



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Bootstrap for '$ python setup.py test' command."""

import os.path
import sys
import unittest
import unittest.runner


def suite():
    test_loader = unittest.TestLoader()
    return test_loader.discover(os.path.dirname(__file__), pattern="test_*.py")


if __name__ == "__main__":
    runner = unittest.runner.TextTestRunner()
    result = runner.run(suite())
    sys.exit(not result.wasSuccessful())



import unittest

from azurefunctions.extensions.base import (
    HttpV2FeatureChecker,
    ModuleTrackerMeta,
    RequestSynchronizer,
    RequestTrackerMeta,
    ResponseLabels,
    ResponseTrackerMeta,
    WebApp,
    WebServer,
)


class TestModuleTrackerMeta(unittest.TestCase):
    def setUp(self):
        # Reset the _module attribute after each test
        ModuleTrackerMeta._module = None
        self.assertFalse(HttpV2FeatureChecker.http_v2_enabled())

    def test_classes_imported_from_same_module(self):
        class TestClass1(metaclass=ModuleTrackerMeta):
            pass

        class TestClass2(metaclass=ModuleTrackerMeta):
            pass

        self.assertEqual(ModuleTrackerMeta.get_module(), __name__)
        self.assertTrue(ModuleTrackerMeta.module_imported())
        self.assertTrue(HttpV2FeatureChecker.http_v2_enabled())

    def test_class_imported_from_a_module(self):
        class TestClass1(metaclass=ModuleTrackerMeta):
            pass

        self.assertEqual(ModuleTrackerMeta.get_module(), __name__)
        self.assertTrue(ModuleTrackerMeta.module_imported())
        self.assertTrue(HttpV2FeatureChecker.http_v2_enabled())

    def test_classes_imported_from_different_modules(self):
        class TestClass1(metaclass=ModuleTrackerMeta):
            __module__ = "module1"

        self.assertEqual(ModuleTrackerMeta.get_module(), "module1")
        self.assertTrue(ModuleTrackerMeta.module_imported())

        with self.assertRaises(Exception) as context:

            class TestClass2(metaclass=ModuleTrackerMeta):
                __module__ = "module2"

        self.assertEqual(
            str(context.exception),
            "Only one web extension package shall be imported, "
            "module1 and module2 are imported",
        )


class TestRequestTrackerMeta(unittest.TestCase):
    class TestRequest1:
        pass

    class TestRequest2:
        pass

    class TestRequest3:
        pass

    class Syncronizer(RequestSynchronizer):
        def sync_route_params(self, request, path_params):
            pass

    def setUp(self):
        # Reset _request_type before each test
        RequestTrackerMeta._request_type = None

    def test_request_type_not_provided(self):
        # Define a class without providing the request_type attribute
        with self.assertRaises(Exception) as context:

            class TestClass(metaclass=RequestTrackerMeta):
                pass

        self.assertEqual(
            str(context.exception), "Request type not provided for class TestClass"
        )

    def test_request_synchronizer_not_provided(self):
        # Define a class without providing the synchronizer attribute
        with self.assertRaises(Exception) as context:

            class TestClass(metaclass=RequestTrackerMeta):
                request_type = self.TestRequest1

        self.assertEqual(
            str(context.exception),
            "Request synchronizer not provided for class TestClass",
        )

    def test_single_request_type(self):
        # Define a class providing a request_type attribute
        class TestClass(metaclass=RequestTrackerMeta):
            request_type = self.TestRequest1
            synchronizer = self.Syncronizer()

        # Ensure the request_type is correctly recorded
        self.assertEqual(RequestTrackerMeta.get_request_type(), self.TestRequest1)
        self.assertTrue(
            isinstance(RequestTrackerMeta.get_synchronizer(), RequestSynchronizer)
        )
        # Ensure check_type returns True for the provided request_type
        self.assertTrue(RequestTrackerMeta.check_type(self.TestRequest1))
        self.assertFalse(RequestTrackerMeta.check_type(self.TestRequest2))

    def test_multiple_request_types_same(self):
        # Define a class providing the same request_type attribute
        class TestClass1(metaclass=RequestTrackerMeta):
            request_type = self.TestRequest1
            synchronizer = self.Syncronizer()

        # Ensure the request_type is correctly recorded
        self.assertEqual(RequestTrackerMeta.get_request_type(), self.TestRequest1)
        self.assertTrue(
            isinstance(RequestTrackerMeta.get_synchronizer(), RequestSynchronizer)
        )
        # Ensure check_type returns True for the provided request_type
        self.assertTrue(RequestTrackerMeta.check_type(self.TestRequest1))

        # Define another class providing the same request_type attribute
        class TestClass2(metaclass=RequestTrackerMeta):
            request_type = self.TestRequest1
            synchronizer = self.Syncronizer()

        # Ensure the request_type remains the same
        self.assertEqual(RequestTrackerMeta.get_request_type(), self.TestRequest1)
        self.assertTrue(
            isinstance(RequestTrackerMeta.get_synchronizer(), RequestSynchronizer)
        )
        # Ensure check_type still returns True for the original request_type
        self.assertTrue(RequestTrackerMeta.check_type(self.TestRequest1))

    def test_multiple_request_types_different(self):
        # Define a class providing a different request_type attribute
        class TestClass1(metaclass=RequestTrackerMeta):
            request_type = self.TestRequest1
            synchronizer = self.Syncronizer()

        # Ensure the request_type is correctly recorded
        self.assertEqual(RequestTrackerMeta.get_request_type(), self.TestRequest1)
        self.assertTrue(
            isinstance(RequestTrackerMeta.get_synchronizer(), RequestSynchronizer)
        )
        # Ensure check_type returns True for the provided request_type
        self.assertTrue(RequestTrackerMeta.check_type(self.TestRequest1))

        # Define another class providing a different request_type attribute
        with self.assertRaises(Exception) as context:

            class TestClass2(metaclass=RequestTrackerMeta):
                request_type = self.TestRequest2

        self.assertEqual(
            str(context.exception),
            f"Only one request type shall be recorded for class TestClass2"
            f" but found {self.TestRequest1} and {self.TestRequest2}",
        )

        # Ensure the request_type remains the same after the exception
        self.assertEqual(RequestTrackerMeta.get_request_type(), self.TestRequest1)
        self.assertTrue(
            isinstance(RequestTrackerMeta.get_synchronizer(), RequestSynchronizer)
        )
        # Ensure check_type still returns True for the original request_type
        self.assertTrue(RequestTrackerMeta.check_type(self.TestRequest1))

    def test_pytype_is_none(self):
        self.assertFalse(RequestTrackerMeta.check_type(None))

    def test_pytype_is_not_class(self):
        self.assertFalse(RequestTrackerMeta.check_type("string"))

    def test_sync_route_params_raises_not_implemented_error(self):
        class MockSyncronizer(RequestSynchronizer):
            def sync_route_params(self, request, path_params):
                super().sync_route_params(request, path_params)

        # Create an instance of RequestSynchronizer
        synchronizer = MockSyncronizer()

        # Ensure that calling sync_route_params raises NotImplementedError
        with self.assertRaises(NotImplementedError):
            synchronizer.sync_route_params(None, None)


class TestResponseTrackerMeta(unittest.TestCase):
    class MockResponse1:
        pass

    class MockResponse2:
        pass

    def test_classes_imported_from_same_module(self):
        class TestResponse1(metaclass=ResponseTrackerMeta):
            label = "test_label_1"
            response_type = self.MockResponse1

        class TestResponse2(metaclass=ResponseTrackerMeta):
            label = "test_label_2"
            response_type = self.MockResponse2

        self.assertEqual(
            ResponseTrackerMeta.get_response_type("test_label_1"), self.MockResponse1
        )
        self.assertEqual(
            ResponseTrackerMeta.get_response_type("test_label_2"), self.MockResponse2
        )
        self.assertIsNone(ResponseTrackerMeta.get_response_type("non_existing_label"))
        self.assertTrue(ResponseTrackerMeta.check_type(self.MockResponse1))
        self.assertTrue(ResponseTrackerMeta.check_type(self.MockResponse2))

    def test_class_imported_from_a_module(self):
        class TestResponse1(metaclass=ResponseTrackerMeta):
            label = "test_label_1"
            response_type = self.MockResponse1

        self.assertEqual(
            ResponseTrackerMeta.get_response_type("test_label_1"), self.MockResponse1
        )
        self.assertIsNone(ResponseTrackerMeta.get_response_type("non_existing_label"))
        self.assertTrue(ResponseTrackerMeta.check_type(self.MockResponse1))
        self.assertFalse(ResponseTrackerMeta.check_type(self.MockResponse2))

    def test_classes_imported_from_different_modules(self):
        class TestResponse1(metaclass=ResponseTrackerMeta):
            __module__ = "module1"
            label = "test_label_1"
            response_type = self.MockResponse1

        with self.assertRaises(Exception) as context:

            class TestResponse2(metaclass=ResponseTrackerMeta):
                __module__ = "module2"
                label = "test_label_1"
                response_type = self.MockResponse2

        self.assertEqual(
            str(context.exception),
            "Only one response type shall be recorded for class TestResponse2 "
            f"but found {self.MockResponse1} and {self.MockResponse2}",
        )

    def test_different_labels(self):
        class TestResponse1(metaclass=ResponseTrackerMeta):
            label = ResponseLabels.STANDARD
            response_type = self.MockResponse1

        class TestResponse2(metaclass=ResponseTrackerMeta):
            label = ResponseLabels.STREAMING
            response_type = self.MockResponse2

        self.assertEqual(
            ResponseTrackerMeta.get_response_type(ResponseLabels.STANDARD),
            self.MockResponse1,
        )
        self.assertEqual(
            ResponseTrackerMeta.get_standard_response_type(), self.MockResponse1
        )
        self.assertEqual(
            ResponseTrackerMeta.get_response_type(ResponseLabels.STREAMING),
            self.MockResponse2,
        )
        self.assertTrue(ResponseTrackerMeta.check_type(self.MockResponse1))
        self.assertTrue(ResponseTrackerMeta.check_type(self.MockResponse2))

    def test_response_label_not_provided(self):
        with self.assertRaises(Exception) as context:

            class TestResponse(metaclass=ResponseTrackerMeta):
                response_type = self.MockResponse1

        self.assertEqual(
            str(context.exception), "Response label not provided for class TestResponse"
        )

    def test_response_type_not_provided(self):
        with self.assertRaises(Exception) as context:

            class TestResponse(metaclass=ResponseTrackerMeta):
                label = "test_label_1"

        self.assertEqual(
            str(context.exception), "Response type not provided for class TestResponse"
        )


class TestWebApp(unittest.TestCase):
    def test_route_and_get_app(self):
        class MockWebApp(WebApp):
            def route(self, func):
                return

            def get_app(self):
                return "MockApp"

        app = MockWebApp()
        self.assertEqual(app.get_app(), "MockApp")

    def test_route_method_raises_not_implemented_error(self):
        class MockWebApp(WebApp):
            def get_app(self):
                pass

            def route(self, func):
                super().route(func)

        with self.assertRaises(NotImplementedError):
            # Create a mock WebApp instance
            mock_web_app = MockWebApp()
            # Call the route method
            mock_web_app.route(None)

    def test_get_app_method_raises_not_implemented_error(self):
        class MockWebApp(WebApp):
            def route(self, func):
                pass

            def get_app(self):
                super().get_app()

        with self.assertRaises(NotImplementedError):
            # Create a mock WebApp instance
            mock_web_app = MockWebApp()
            # Call the get_app method
            mock_web_app.get_app()


class TestWebServer(unittest.TestCase):
    def test_web_server_initialization(self):
        class MockWebApp(WebApp):
            def route(self, func):
                pass

            def get_app(self):
                return "MockApp"

        class MockWebServer(WebServer):
            async def serve(self):
                pass

        mock_web_app = MockWebApp()
        server = MockWebServer("localhost", 8080, mock_web_app)
        self.assertEqual(server.hostname, "localhost")
        self.assertEqual(server.port, 8080)
        self.assertEqual(server.web_app, "MockApp")


class TestHttpV2Enabled(unittest.TestCase):
    def test_http_v2_enabled(self):
        ModuleTrackerMeta._module = None

        class MockClass(metaclass=ModuleTrackerMeta):
            pass

        MockClass()

        self.assertTrue(HttpV2FeatureChecker.http_v2_enabled())


class TestResponseLabels(unittest.TestCase):
    def test_enum_values(self):
        self.assertEqual(ResponseLabels.STANDARD.value, "standard")
        self.assertEqual(ResponseLabels.STREAMING.value, "streaming")
        self.assertEqual(ResponseLabels.FILE.value, "file")
        self.assertEqual(ResponseLabels.HTML.value, "html")
        self.assertEqual(ResponseLabels.JSON.value, "json")
        self.assertEqual(ResponseLabels.ORJSON.value, "orjson")
        self.assertEqual(ResponseLabels.PLAIN_TEXT.value, "plain_text")
        self.assertEqual(ResponseLabels.REDIRECT.value, "redirect")
        self.assertEqual(ResponseLabels.UJSON.value, "ujson")
        self.assertEqual(ResponseLabels.INT.value, "int")
        self.assertEqual(ResponseLabels.FLOAT.value, "float")
        self.assertEqual(ResponseLabels.STR.value, "str")
        self.assertEqual(ResponseLabels.LIST.value, "list")
        self.assertEqual(ResponseLabels.DICT.value, "dict")
        self.assertEqual(ResponseLabels.BOOL.value, "bool")
        self.assertEqual(ResponseLabels.PYDANTIC.value, "pydantic")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import unittest
from typing import List, Mapping
from unittest.mock import patch

from azurefunctions.extensions.base import meta, sdkType


class TestMeta(unittest.TestCase):
    # Test Datum class
    def test_datum_single_level_python_value(self):
        datum: Mapping[str, meta.Datum] = meta.Datum(value=None, type="int")
        self.assertEqual(datum.python_value, None)
        self.assertEqual(datum.python_type, type(None))

        datum = meta.Datum(value=1, type=None)
        self.assertEqual(datum.python_value, None)
        self.assertEqual(datum.python_type, type(None))

        datum = meta.Datum(value=b"awesome bytes", type="bytes")
        self.assertEqual(datum.python_value, b"awesome bytes")
        self.assertEqual(datum.python_type, bytes)

        datum = meta.Datum(value="awesome string", type="string")
        self.assertEqual(datum.python_value, "awesome string")
        self.assertEqual(datum.python_type, str)

        datum = meta.Datum(value=42, type="int")
        self.assertEqual(datum.python_value, 42)
        self.assertEqual(datum.python_type, int)

        datum = meta.Datum(value=43.2103, type="double")
        self.assertEqual(datum.python_value, 43.2103)
        self.assertEqual(datum.python_type, float)

        datum = meta.Datum(value="other", type="other")
        self.assertEqual(datum.python_value, "other")

    def test_datum_collections_python_value(self):
        class DatumCollectionString:
            def __init__(self, *args: List[str]):
                self.string = args

        datum = meta.Datum(
            value=DatumCollectionString("string 1", "string 2"),
            type="collection_string",
        )
        self.assertListEqual(datum.python_value, ["string 1", "string 2"])
        self.assertEqual(datum.python_type, list)

        class DatumCollectionBytes:
            def __init__(self, *args: List[bytes]):
                self.bytes = args

        datum = meta.Datum(
            value=DatumCollectionBytes(b"bytes 1", b"bytes 2"), type="collection_bytes"
        )
        self.assertListEqual(datum.python_value, [b"bytes 1", b"bytes 2"])
        self.assertEqual(datum.python_type, list)

        class DatumCollectionDouble:
            def __init__(self, *args: List[bytes]):
                self.double = args

        datum = meta.Datum(
            value=DatumCollectionDouble(43.2103, 45.601), type="collection_double"
        )
        self.assertListEqual(datum.python_value, [43.2103, 45.601])
        self.assertEqual(datum.python_type, list)

        class DatumCollectionSint64:
            def __init__(self, *args: List[int]):
                self.sint64 = args

        datum = meta.Datum(
            value=DatumCollectionSint64(1234567, 8901234), type="collection_sint64"
        )
        self.assertListEqual(datum.python_value, [1234567, 8901234])
        self.assertEqual(datum.python_type, list)

    def test_datum_json_python_value(self):
        # None
        datum = meta.Datum(value="null", type="json")
        self.assertEqual(datum.python_value, None)
        self.assertEqual(datum.python_type, type(None))

        # Int
        datum = meta.Datum(value="123", type="json")
        self.assertEqual(datum.python_value, 123)
        self.assertEqual(datum.python_type, int)

        # Float
        datum = meta.Datum(value="456.789", type="json")
        self.assertEqual(datum.python_value, 456.789)
        self.assertEqual(datum.python_type, float)

        # String
        datum = meta.Datum(value='"string in json"', type="json")
        self.assertEqual(datum.python_value, "string in json")
        self.assertEqual(datum.python_type, str)

        # List
        datum = meta.Datum(value='["a", "b", "c"]', type="json")
        self.assertListEqual(datum.python_value, ["a", "b", "c"])
        self.assertEqual(datum.python_type, list)

        # Object
        datum = meta.Datum(value='{"name": "awesome", "value": "cool"}', type="json")
        self.assertDictEqual(datum.python_value, {"name": "awesome", "value": "cool"})
        self.assertEqual(datum.python_type, dict)

        # Should ignore Newlines and Spaces
        datum = meta.Datum(
            value='{ "name" : "awesome",\n "value":  "cool"\n}', type="json"
        )
        self.assertDictEqual(datum.python_value, {"name": "awesome", "value": "cool"})
        self.assertEqual(datum.python_type, dict)

    def test_equals(self):
        str_datum = meta.Datum(value="awesome string", type="string")
        str_datum_copy = meta.Datum(value="awesome string", type="string")
        str_datum_wrong_copy = meta.Datum(value="not awesome string", type="string")
        self.assertFalse(str_datum.__eq__(dict))
        self.assertTrue(str_datum.__eq__(str_datum_copy))
        self.assertFalse(str_datum.__eq__(str_datum_wrong_copy))

    def test_hash(self):
        str_datum = meta.Datum(value="awesome string", type="string")
        datum_hash = str_datum.__hash__()
        self.assertIsInstance(datum_hash, int)

    def test_repr(self):
        str_datum = meta.Datum(value="awesome", type="string")
        self.assertEqual(str_datum.__repr__(), "<Datum string 'awesome'>")

        long_str_datum = meta.Datum(value="awesome string", type="string")
        self.assertEqual(long_str_datum.__repr__(), "<Datum string 'awesome s...>")

    def test_registry(self):
        registry = meta.get_binding_registry()
        self.assertIsInstance(registry, type(meta._ConverterMeta))
        self.assertIsNone(registry.get("test"))

        class MockIndexedFunction:
            _bindings = {}
            _trigger = None

        self.assertEqual(registry.get_raw_bindings(MockIndexedFunction, []), ([], {}))

        self.assertFalse(registry.check_supported_type(None))
        self.assertFalse(registry.check_supported_type("hello"))
        self.assertTrue(registry.check_supported_type(sdkType.SdkType))

        self.assertFalse(registry.has_trigger_support(MockIndexedFunction))

    def test_decode_typed_data(self):
        # Case 1: data is None
        self.assertIsNone(
            meta._BaseConverter._decode_typed_data(data=None, python_type=str)
        )

        # Case 2: data.type is model_binding_data
        datum_mbd = meta.Datum(value="{}", type="model_binding_data")
        self.assertEqual(
            meta._BaseConverter._decode_typed_data(datum_mbd, python_type=str), "{}"
        )

        # Case 3: data.type is None
        datum_none = meta.Datum(value="{}", type=None)
        self.assertIsNone(
            meta._BaseConverter._decode_typed_data(datum_none, python_type=str)
        )

        # Case 4: data.type is unsupported
        datum_unsupp = meta.Datum(value="{}", type=dict)
        with self.assertRaises(ValueError):
            meta._BaseConverter._decode_typed_data(datum_unsupp, python_type=str)

        # Case 5: can't coerce
        datum_coerce_fail = meta.Datum(value="{}", type="model_binding_data")
        with self.assertRaises(ValueError):
            meta._BaseConverter._decode_typed_data(
                datum_coerce_fail, python_type=(tuple, list, dict)
            )

        # Case 6: attempt coerce & fail
        datum_attempt_coerce = meta.Datum(value=1, type="model_binding_data")
        with self.assertRaises(ValueError):
            meta._BaseConverter._decode_typed_data(
                datum_attempt_coerce, python_type=dict
            )

        # Case 7: attempt to coerce and pass
        datum_coerce_pass = meta.Datum(value=1, type="model_binding_data")
        self.assertEqual(
            meta._BaseConverter._decode_typed_data(datum_coerce_pass, python_type=str),
            "1",
        )

    def test_decode_trigger_metadata_field(self):
        datum_mbd = meta.Datum(value="{}", type="model_binding_data")
        mock_trigger_metadata = {"key": datum_mbd}

        self.assertIsNone(
            meta._BaseConverter._decode_trigger_metadata_field(
                trigger_metadata=mock_trigger_metadata, field="fakeKey", python_type=str
            )
        )

        self.assertEqual(
            meta._BaseConverter._decode_trigger_metadata_field(
                trigger_metadata=mock_trigger_metadata, field="key", python_type=str
            ),
            "{}",
        )

    @patch(
        "azurefunctions.extensions.base.meta." "InConverter.__abstractmethods__", set()
    )
    def test_in_converter(self):
        class MockInConverter(meta.InConverter, binding="test1"):
            _sample = ""

        mock_converter = MockInConverter()
        self.assertIsNone(mock_converter.check_input_type_annotation(pytype=str))

        with self.assertRaises(NotImplementedError):
            mock_converter.decode(data=None, trigger_metadata={})

        self.assertFalse(mock_converter.has_implicit_output())

    @patch(
        "azurefunctions.extensions.base.meta." "OutConverter.__abstractmethods__", set()
    )
    def test_out_converter(self):
        class MockOutConverter(meta.OutConverter, binding="test2"):
            _sample = ""

        mock_converter = MockOutConverter()
        mock_converter.check_output_type_annotation(pytype=str)

        with self.assertRaises(NotImplementedError):
            mock_converter.encode(obj=None, expected_type=None)

    def test_get_registry(self):
        registry = meta.get_binding_registry()
        self.assertEqual(registry, meta._ConverterMeta)

    @patch(
        "azurefunctions.extensions.base.meta." "OutConverter.__abstractmethods__", set()
    )
    def test_converter_meta(self):
        class BindingNoneConverter(meta.OutConverter, binding=None):
            _sample = ""

        registry = meta.get_binding_registry()
        self.assertEqual(len(registry._bindings), 0)

        class BindingBlobConverter(meta.OutConverter, binding="blob"):
            _sample = ""

        registry = meta.get_binding_registry()
        self.assertEqual(len(registry._bindings), 1)
        self.assertIsNotNone(registry._bindings.get("blob"))
        self.assertEqual(registry._bindings.get("blob"), BindingBlobConverter)

        with self.assertRaises(RuntimeError):

            class BindingBlob2Converter(meta.OutConverter, binding="blob"):
                _sample = ""

        registry = meta.get_binding_registry()
        self.assertEqual(len(registry._bindings), 1)
        self.assertIsNotNone(registry._bindings.get("blob"))
        self.assertEqual(registry._bindings.get("blob"), BindingBlobConverter)

        class BindingServiceBusConverter(
            meta.OutConverter, binding="serviceBus", trigger="serviceBusTrigger"
        ):
            _sample = ""

        registry = meta.get_binding_registry()
        self.assertEqual(len(registry._bindings), 3)
        self.assertIsNotNone(registry._bindings.get("serviceBus"))
        self.assertEqual(
            registry._bindings.get("serviceBus"), BindingServiceBusConverter
        )
        self.assertIsNotNone(registry._bindings.get("serviceBusTrigger"))
        self.assertEqual(
            registry._bindings.get("serviceBusTrigger"), BindingServiceBusConverter
        )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from abc import abstractmethod


class SdkType:
    def __init__(self, *, data: dict = None):
        self._data = data or {}

    @abstractmethod
    def get_sdk_type(self):
        pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest
from abc import ABC

from azurefunctions.extensions.base import meta, sdkType, utils


class MockParamTypeInfo:
    def __init__(self, binding_name: str, pytype: type):
        self.binding_name = binding_name
        self.pytype = pytype


class MockFunction(ABC):
    def __init__(self, bindings: utils.Binding):
        self._bindings = bindings


class MockInitParams(utils.Binding):
    def __init__(self, name, direction, data_type, type, init_params, path=None):
        self.type = "blob"
        self.name = name
        self._direction = direction
        self._data_type = data_type
        self._dict = {
            "direction": self._direction,
            "dataType": self._data_type,
            "type": self.type,
        }
        self.init_params = init_params
        self.path = path


class TestUtils(unittest.TestCase):
    # Test Utils class
    def test_get_dict_repr_sdk(self):
        # Create mock blob
        meta._ConverterMeta._bindings = {"blob"}

        # Create test binding
        mock_blob = utils.Binding(
            name="client",
            direction=utils.BindingDirection.IN,
            data_type=None,
            type="blob",
        )

        # Create test input_types dict
        mock_input_types = {
            "client": MockParamTypeInfo(
                binding_name="blobTrigger", pytype=sdkType.SdkType
            )
        }

        # Create test indexed_function
        mock_indexed_functions = MockFunction(bindings=[mock_blob])

        dict_repr, logs = utils.get_raw_bindings(
            mock_indexed_functions, mock_input_types
        )
        self.assertEqual(
            dict_repr,
            [
                '{"direction": "IN", '
                '"type": "blob", '
                '"properties": '
                '{"SupportsDeferredBinding": true}}'
            ],
        )

        self.assertEqual(logs, {"client": {sdkType.SdkType: "True"}})

    def test_get_dict_repr_non_sdk(self):
        # Create mock blob
        meta._ConverterMeta._bindings = {"blob"}

        # Create test binding
        mock_blob = utils.Binding(
            name="blob",
            direction=utils.BindingDirection.IN,
            data_type=None,
            type="blob",
        )

        # Create test input_types dict
        mock_input_types = {
            "blob": MockParamTypeInfo(binding_name="blobTrigger", pytype=bytes)
        }

        # Create test indexed_function
        mock_indexed_functions = MockFunction(bindings=[mock_blob])

        dict_repr, logs = utils.get_raw_bindings(
            mock_indexed_functions, mock_input_types
        )
        self.assertEqual(
            dict_repr,
            [
                '{"direction": "IN", '
                '"type": "blob", '
                '"properties": '
                '{"SupportsDeferredBinding": false}}'
            ],
        )
        self.assertEqual(logs, {"blob": {bytes: "False"}})

    def test_get_dict_repr_binding_name_none(self):
        # Create mock blob
        meta._ConverterMeta._bindings = {"blob"}

        # Create test binding
        mock_blob = utils.Binding(
            name="blob",
            direction=utils.BindingDirection.IN,
            data_type=None,
            type="blob",
        )

        mock_http = utils.Binding(
            name="$return",
            direction=utils.BindingDirection.OUT,
            data_type=None,
            type="httpResponse",
        )

        # Create test input_types dict
        mock_input_types = {
            "blob": MockParamTypeInfo(binding_name="blobTrigger", pytype=bytes)
        }

        # Create test indexed_function
        mock_indexed_functions = MockFunction(bindings=[mock_blob, mock_http])

        dict_repr, logs = utils.get_raw_bindings(
            mock_indexed_functions, mock_input_types
        )
        self.assertEqual(
            dict_repr,
            [
                '{"direction": "IN", "type": "blob", '
                '"properties": {"SupportsDeferredBinding": false}}',
                '{"direction": "OUT", "type": "httpResponse", '
                '"properties": {"SupportsDeferredBinding": false}}',
            ],
        )
        self.assertEqual(logs, {"$return": {None: "False"}, "blob": {bytes: "False"}})

    def test_get_dict_repr_init_params(self):
        # Create mock blob
        meta._ConverterMeta._bindings = {"blob"}

        # Create test binding
        mock_blob = MockInitParams(
            name="client",
            direction=utils.BindingDirection.IN,
            data_type=None,
            type="blob",
            init_params=["test", "type", "direction"],
        )

        # Create test input_types dict
        mock_input_types = {
            "client": MockParamTypeInfo(
                binding_name="blobTrigger", pytype=sdkType.SdkType
            )
        }

        # Create test indexed_function
        mock_indexed_functions = MockFunction(bindings=[mock_blob])

        dict_repr, logs = utils.get_raw_bindings(
            mock_indexed_functions, mock_input_types
        )
        self.assertEqual(
            dict_repr,
            [
                '{"direction": "IN", '
                '"type": "blob", "properties": '
                '{"SupportsDeferredBinding": true}}'
            ],
        )

        self.assertEqual(logs, {"client": {sdkType.SdkType: "True"}})

    def test_get_dict_repr_clean_nones(self):
        # Create mock blob
        meta._ConverterMeta._bindings = {"blob"}

        # Create test binding
        mock_blob = MockInitParams(
            name="client",
            direction=utils.BindingDirection.IN,
            data_type=None,
            type="blob",
            path=None,
            init_params=["test", "type", "direction", "path"],
        )

        # Create test input_types dict
        mock_input_types = {
            "client": MockParamTypeInfo(
                binding_name="blobTrigger", pytype=sdkType.SdkType
            )
        }

        # Create test indexed_function
        mock_indexed_functions = MockFunction(bindings=[mock_blob])

        dict_repr, logs = utils.get_raw_bindings(
            mock_indexed_functions, mock_input_types
        )
        self.assertEqual(
            dict_repr,
            [
                '{"direction": "IN", '
                '"type": "blob", "properties": '
                '{"SupportsDeferredBinding": true}}'
            ],
        )

        self.assertEqual(logs, {"client": {sdkType.SdkType: "True"}})

    def test_binding_data_type(self):
        mock_blob = utils.Binding(
            name="blob",
            direction=utils.BindingDirection.IN,
            data_type=None,
            type="blob",
        )
        self.assertIsNone(mock_blob.data_type)

        mock_data_type = utils.Binding(
            name="blob",
            direction=utils.BindingDirection.IN,
            data_type=utils.DataType.STRING,
            type="blob",
        )
        self.assertEqual(mock_data_type.data_type, 1)

    def test_to_camel_case(self):
        test_str = ""
        self.assertRaises(ValueError, utils.to_camel_case, test_str)

        test_str = "1iAmNotAWord"
        self.assertRaises(ValueError, utils.to_camel_case, test_str)

        test_str = utils.to_camel_case("string_in_correct_format")
        self.assertEqual(test_str, "stringInCorrectFormat")

    def test_is_snake_case(self):
        test_str = "foo_bar_baz"
        self.assertTrue(utils.is_snake_case(test_str))

        test_str = "foo"
        self.assertFalse(utils.is_snake_case(test_str))

    def test_is_word(self):
        test_str = "1foo"
        self.assertFalse(utils.is_word(test_str))

        test_str = "foo_"
        self.assertFalse(utils.is_word(test_str))

        test_str = "foo"
        self.assertTrue(utils.is_word(test_str))

    def test_snake_case_to_camel_case_multi(self):
        self.assertEqual(utils.to_camel_case("data_type"), "dataType")

    def test_snake_case_to_camel_case_trailing_underscore(self):
        self.assertEqual(utils.to_camel_case("data_type_"), "dataType")

    def test_snake_case_to_camel_case_leading_underscore(self):
        self.assertEqual(utils.to_camel_case("_dataType"), "Datatype")

    def test_snake_case_to_camel_case_single(self):
        self.assertEqual(utils.to_camel_case("dataType"), "dataType")

    def test_snake_case_to_camel_case_empty_str(self):
        with self.assertRaises(ValueError) as err:
            utils.to_camel_case("")
        self.assertEqual(
            err.exception.args[0], "Please ensure arg name  is not " "empty!"
        )

    def test_snake_case_to_camel_case_none(self):
        with self.assertRaises(ValueError) as err:
            utils.to_camel_case(None)
        self.assertEqual(
            err.exception.args[0], "Please ensure arg name None is not " "empty!"
        )

    def test_snake_case_to_camel_case_not_one_word_nor_snake_case(self):
        with self.assertRaises(ValueError) as err:
            utils.to_camel_case("data-type")
        self.assertEqual(
            err.exception.args[0],
            "Please ensure data-type is a word or snake case "
            "string with underscore as separator.",
        )

    def test_is_snake_case_letters_only(self):
        self.assertTrue(utils.is_snake_case("dataType_foo"))

    def test_is_snake_case_lowercase_with_digit(self):
        self.assertTrue(utils.is_snake_case("data_type_233"))

    def test_is_snake_case_uppercase_with_digit(self):
        self.assertTrue(utils.is_snake_case("Data_Type_233"))

    def test_is_snake_case_leading_digit(self):
        self.assertFalse(utils.is_snake_case("233_Data_Type_233"))

    def test_is_snake_case_no_separator(self):
        self.assertFalse(utils.is_snake_case("DataType233"))

    def test_is_snake_case_invalid_separator(self):
        self.assertFalse(utils.is_snake_case("Data-Type-233"))

    def test_is_word_letters_only(self):
        self.assertTrue(utils.is_word("dataType"))

    def test_is_word_letters_with_digits(self):
        self.assertTrue(utils.is_word("dataType233"))

    def test_is_word_leading_digits(self):
        self.assertFalse(utils.is_word("233dataType"))

    def test_is_word_invalid_symbol(self):
        self.assertFalse(utils.is_word("233!dataType"))

    def test_clean_nones_none(self):
        self.assertEqual(utils.BuildDictMeta.clean_nones(None), None)

    def test_clean_nones_nested(self):
        self.assertEqual(
            utils.BuildDictMeta.clean_nones(
                {
                    "hello": None,
                    "hello2": [
                        "dummy1",
                        None,
                        "dummy2",
                        ["dummy3", None],
                        {"hello3": None},
                    ],
                    "hello4": {"dummy5": "pass1", "dummy6": None},
                }
            ),
            {
                "hello2": ["dummy1", "dummy2", ["dummy3"], {}],
                "hello4": {"dummy5": "pass1"},
            },  # NoQA
        )

    def test_add_to_dict_no_args(self):
        with self.assertRaises(ValueError) as err:

            @utils.BuildDictMeta.add_to_dict
            def dummy():
                pass

            dummy()

        self.assertEqual(
            err.exception.args[0],
            "dummy has no args. Please ensure func is an object " "method.",
        )

    def test_add_to_dict_valid(self):
        class TestDict:
            @utils.BuildDictMeta.add_to_dict
            def __init__(self, arg1, arg2, **kwargs):
                self.arg1 = arg1
                self.arg2 = arg2

        test_obj = TestDict("val1", "val2", dummy1="dummy1", dummy2="dummy2")

        self.assertCountEqual(
            getattr(test_obj, "init_params"),
            {"self", "arg1", "arg2", "kwargs", "dummy1", "dummy2"},
        )
        self.assertEqual(getattr(test_obj, "arg1", None), "val1")
        self.assertEqual(getattr(test_obj, "arg2", None), "val2")
        self.assertEqual(getattr(test_obj, "dummy1", None), "dummy1")
        self.assertEqual(getattr(test_obj, "dummy2", None), "dummy2")

    def test_add_to_dict_duplicate(self):
        class TestDict:
            @utils.BuildDictMeta.add_to_dict
            def __init__(self, arg1, arg2, **kwargs):
                self.arg1 = arg1
                self.arg2 = arg2
                self.arg3 = arg1

        test_obj = TestDict("val1", "val2", arg3="dummy1")

        self.assertCountEqual(
            getattr(test_obj, "init_params"), {"self", "arg1", "arg2", "kwargs", "arg3"}
        )
        self.assertEqual(getattr(test_obj, "arg1", None), "val1")
        self.assertEqual(getattr(test_obj, "arg2", None), "val2")
        self.assertEqual(getattr(test_obj, "arg3", None), "val1")

    def test_build_dict_meta(self):
        class TestBuildDict(metaclass=utils.BuildDictMeta):
            def __init__(self, arg1, arg2):
                pass

            def get_dict_repr(self):
                return {"hello": None, "world": ["dummy", None]}

        test_obj = TestBuildDict("val1", "val2")

        self.assertCountEqual(
            getattr(test_obj, "init_params"), {"self", "arg1", "arg2"}
        )
        self.assertEqual(test_obj.get_dict_repr(), {"world": ["dummy"]})



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import abc
import inspect
import json
from typing import Any, Dict, Mapping, Optional, Tuple, Union

from . import sdkType, utils


class Datum:
    def __init__(self, value: Any, type: Optional[str]):
        self.value: Any = value
        self.type: Optional[str] = type

    @property
    def python_value(self) -> Any:
        if self.value is None or self.type is None:
            return None
        elif self.type in ("bytes", "string", "int", "double"):
            return self.value
        elif self.type == "json":
            return json.loads(self.value)
        elif self.type == "collection_string":
            return [v for v in self.value.string]
        elif self.type == "collection_bytes":
            return [v for v in self.value.bytes]
        elif self.type == "collection_double":
            return [v for v in self.value.double]
        elif self.type == "collection_sint64":
            return [v for v in self.value.sint64]
        else:
            return self.value

    @property
    def python_type(self) -> type:
        return type(self.python_value)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return self.value == other.value and self.type == other.type

    def __hash__(self):
        return hash((type(self), (self.value, self.type)))

    def __repr__(self):
        val_repr = repr(self.value)
        if len(val_repr) > 10:
            val_repr = val_repr[:10] + "..."
        return "<Datum {} {}>".format(self.type, val_repr)


class _ConverterMeta(abc.ABCMeta):

    _bindings: Dict[str, type] = {}

    def __new__(
        mcls, name, bases, dct, *, binding: Optional[str], trigger: Optional[str] = None
    ):
        cls = super().__new__(mcls, name, bases, dct)
        cls._trigger = trigger  # type: ignore
        if binding is None:
            return cls

        if binding in mcls._bindings:
            raise RuntimeError(
                f"cannot register a converter for {binding!r} binding: "
                f"another converter for this binding has already been "
                f"registered"
            )

        mcls._bindings[binding] = cls
        if trigger is not None:
            mcls._bindings[trigger] = cls

        return cls

    @classmethod
    def get(cls, binding_name):
        return cls._bindings.get(binding_name)

    @classmethod
    def get_raw_bindings(cls, indexed_function, input_types):
        return utils.get_raw_bindings(indexed_function, input_types)

    @classmethod
    def check_supported_type(cls, subclass: type) -> bool:
        if subclass is not None and inspect.isclass(subclass):
            return issubclass(subclass, sdkType.SdkType)
        return False

    def has_trigger_support(cls) -> bool:
        return cls._trigger is not None  # type: ignore


class _BaseConverter(metaclass=_ConverterMeta, binding=None):

    @classmethod
    def _decode_typed_data(
        cls,
        data: Datum,
        *,
        python_type: Union[type, Tuple[type, ...]],
        context: str = "data",
    ) -> Any:
        if data is None:
            return None

        data_type = data.type
        if data_type == "model_binding_data":
            result = data.value
        elif data_type is None:
            return None
        else:
            raise ValueError(f"unsupported type of {context}: {data_type}")

        if not isinstance(result, python_type):
            if isinstance(python_type, (tuple, list, dict)):
                raise ValueError(
                    f"unexpected value type in {context}: "
                    f"{type(result).__name__}, expected one of: "
                    f'{", ".join(t.__name__ for t in python_type)}'
                )
            else:
                try:
                    # Try coercing into the requested type
                    result = python_type(result)
                except (TypeError, ValueError) as e:
                    raise ValueError(
                        f"cannot convert value of {context} into "
                        f"{python_type.__name__}: {e}"
                    ) from None

        return result

    @classmethod
    def _decode_trigger_metadata_field(
        cls,
        trigger_metadata: Mapping[str, Datum],
        field: str,
        *,
        python_type: Union[type, Tuple[type, ...]],
    ) -> Any:
        data = trigger_metadata.get(field)
        if data is None:
            return None
        else:
            return cls._decode_typed_data(
                data,
                python_type=python_type,
                context=f"field {field!r} in trigger metadata",
            )


class InConverter(_BaseConverter, binding=None):

    @classmethod
    @abc.abstractmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        pass

    @classmethod
    @abc.abstractmethod
    def decode(cls, data: Datum, *, trigger_metadata) -> Any:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def has_implicit_output(cls) -> bool:
        return False


class OutConverter(_BaseConverter, binding=None):

    @classmethod
    @abc.abstractmethod
    def check_output_type_annotation(cls, pytype: type) -> bool:
        pass

    @classmethod
    @abc.abstractmethod
    def encode(cls, obj: Any, *, expected_type: Optional[type]) -> Optional[Datum]:
        raise NotImplementedError


def get_binding_registry():
    return _ConverterMeta



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import unittest

from azurefunctions.extensions.base import sdkType


class TestSdkType(unittest.TestCase):

    def test_init(self):
        data_populated = sdkType.SdkType(data={"key": "value"})
        self.assertEqual(data_populated._data, {"key": "value"})

        data_empty = sdkType.SdkType()
        self.assertEqual(data_empty._data, {})

    def test_get_sdk_type(self):
        class MockSdkType(sdkType.SdkType):
            _sample = ""

        mock_type = MockSdkType()
        self.assertIsNone(mock_type.get_sdk_type())



import abc
import inspect
from abc import abstractmethod
from enum import Enum
from typing import Callable

base_extension_module = __name__


# Base extension pkg
class ModuleTrackerMeta(type):
    _module = None

    def __new__(cls, name, bases, dct, **kwargs):
        new_class = super().__new__(cls, name, bases, dct)
        new_module = dct.get("__module__")
        if new_module != base_extension_module:
            if cls._module is None:
                cls._module = new_module
            elif cls._module != new_module:
                raise Exception(
                    f"Only one web extension package shall be imported, "
                    f"{cls._module} and {new_module} are imported"
                )
        return new_class

    @classmethod
    def get_module(cls):
        return cls._module

    @classmethod
    def module_imported(cls):
        return cls._module is not None


class RequestTrackerMeta(type):
    _request_type = None
    _synchronizer: None

    def __new__(cls, name, bases, dct, **kwargs):
        new_class = super().__new__(cls, name, bases, dct)

        request_type = dct.get("request_type")

        if request_type is None:
            raise TypeError(f"Request type not provided for class {name}")

        if cls._request_type is not None and cls._request_type != request_type:
            raise TypeError(
                f"Only one request type shall be recorded for class {name} "
                f"but found {cls._request_type} and {request_type}"
            )
        cls._request_type = request_type
        cls._synchronizer = dct.get("synchronizer")

        if cls._synchronizer is None:
            raise TypeError(f"Request synchronizer not provided for class {name}")

        return new_class

    @classmethod
    def get_request_type(cls):
        return cls._request_type

    @classmethod
    def get_synchronizer(cls):
        return cls._synchronizer

    @classmethod
    def check_type(cls, pytype: type) -> bool:
        if pytype is not None and inspect.isclass(pytype):
            return cls._request_type is not None and issubclass(
                pytype, cls._request_type
            )
        return False


class RequestSynchronizer(abc.ABC):
    @abstractmethod
    def sync_route_params(self, request, path_params):
        raise NotImplementedError()


class ResponseTrackerMeta(type):
    _response_types = {}

    def __new__(cls, name, bases, dct, **kwargs):
        new_class = super().__new__(cls, name, bases, dct)

        label = dct.get("label")
        response_type = dct.get("response_type")

        if label is None:
            raise TypeError(f"Response label not provided for class {name}")
        if response_type is None:
            raise TypeError(f"Response type not provided for class {name}")
        if (
            cls._response_types.get(label) is not None
            and cls._response_types.get(label) != response_type
        ):
            raise TypeError(
                f"Only one response type shall be recorded for class {name} "
                f"but found {cls._response_types.get(label)} and {response_type}"
            )

        cls._response_types[label] = response_type

        return new_class

    @classmethod
    def get_standard_response_type(cls):
        return cls.get_response_type(ResponseLabels.STANDARD)

    @classmethod
    def get_response_type(cls, label):
        return cls._response_types.get(label)

    @classmethod
    def check_type(cls, pytype: type) -> bool:
        if pytype is not None and inspect.isclass(pytype):
            return cls._response_types is not None and any(
                issubclass(pytype, response_type)
                for response_type in cls._response_types.values()
            )
        return False


class WebApp(metaclass=ModuleTrackerMeta):
    @abstractmethod
    def route(self, func: Callable):
        raise NotImplementedError()

    @abstractmethod
    def get_app(self):
        raise NotImplementedError()  # pragma: no cover


class WebServer(metaclass=ModuleTrackerMeta):
    def __init__(self, hostname, port, web_app: WebApp):
        self.hostname = hostname
        self.port = port
        self.web_app = web_app.get_app()

    @abstractmethod
    async def serve(self):
        raise NotImplementedError()  # pragma: no cover


class HttpV2FeatureChecker:
    @staticmethod
    def http_v2_enabled():
        return ModuleTrackerMeta.module_imported()


class ResponseLabels(Enum):
    STANDARD = "standard"
    STREAMING = "streaming"
    FILE = "file"
    HTML = "html"
    JSON = "json"
    ORJSON = "orjson"
    PLAIN_TEXT = "plain_text"
    REDIRECT = "redirect"
    UJSON = "ujson"
    INT = "int"
    FLOAT = "float"
    STR = "str"
    LIST = "list"
    DICT = "dict"
    BOOL = "bool"
    PYDANTIC = "pydantic"
